/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file conn.cpp
 *
 * Implements the cconn class, which is the object for an open connection to
 * another P4 client. For more information, see conn.h.
 *
 * TODO: - return an error if message is too big (in api call?)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "lka.h"
#include "conn_mgr.h"
#include "conn.h"
#include "msg_factory.h"
#include "msg.h"
#include "debug.h"
#include "p4error.h"

#ifdef _WIN32
#include <Winsock2.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>    // for close
#endif

#include <cassert>
#include <cerrno>

/*****************************************************************************
 * INITIALIZE STATIC MEMBERS
 *****************************************************************************/
const uint cconn::kconn_wait_until_ready_timeout_msecs =
  csettings::instance().get<uint>("conn_wait_until_ready_timeout_msecs");

/*****************************************************************************
 * PUBLIC FUNCTIONS
 *****************************************************************************/
ostream &operator<< (ostream &ostr, const cconn &conn)
{
  ostr << "conn: " << " ";

  // print status
  switch (conn.status) {
  case cconn::eopening:           ostr << "opening"; break;
  case cconn::epreamble_wait:     ostr << "preamble_wait"; break;
  case cconn::epreamble_ack_wait: ostr << "preamble_ack_wait"; break;
  case cconn::econnected:         ostr << "connected"; break;
  case cconn::efailed:            ostr << "failed"; break;
  case cconn::eclosed:            ostr << "closed"; break;
  default:                        assert(false);
  }

  ostr << conn.peer_username;

  if (!conn.permanent)
    ostr << " (transient!)";

  ostr << endl;
  return ostr;
}


uint cconn::wait_until_ready() const
{
  // fill in timeout for select
  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = kconn_wait_until_ready_timeout_msecs * 1000;

  // ugly spin lock to avoid a lost wakeup that would occur with a condition
  // TODO: replace this with a pthread_cond. no lost wakeup problem.
  while (status == eopening || status == epreamble_wait ||
         status == epreamble_ack_wait)
    select(0, 0, 0, 0, &timeout);

  if (status == econnected)
    return kerr_success;
  else
    return kerr_connect_failed;
}


cconn::~cconn()
{
  cdebug::instance() << acquire("conn_debug") << "dtor()" << release();

  if (msg)
    delete msg;
}


/*****************************************************************************
 * PROTECTED METHODS
 *****************************************************************************/

uint cconn::send(cmsg &msg)
{
//  wait_until_ready();
  if (status != econnected)
    return kerr_not_sent;

  cdebug::instance() << acquire("conn_debug")
                     << "send() waiting till ready " << msg << " to "
                     << peer_username
                     << release();
  wait_until_ready();
  assert(status == econnected);

  cdebug::instance() << acquire("conn_debug") << "send() with msg " << msg
                     << " to " << peer_username
                     << release();

  // convert byte order
  if (msg.byte_order == cmsg::ehost)
    msg.hton();
  assert(msg.byte_order == cmsg::enetwork);

  uint ret = send_all(&msg.hdr, sizeof(msg.hdr));
  if (ret != kerr_success) {
    cdebug::instance() << acquire("conn_debug")
                       << "couldn't send!"
                       << release();
    return ret;
  }

  if (msg.data) {
    ret = send_all(msg.data, msg.host_data_len);
    if (ret != kerr_success) {
      cdebug::instance() << acquire("conn_debug")
                         << "couldn't send data!"
                         << release();
      return ret;
    }
  }
  
  if (msg.hdr.is_type(kp4_connect))  // detect if outgoing connection msg
    make_permanent();

  return kerr_success;
}


uint cconn::receive()
{
  if (status == eopening)
    return kerr_success;
  if (status == efailed || status == eclosed)
    return kerr_couldnt_receive;


  // are we still doing the preamble?
  if (status == epreamble_wait || status == epreamble_ack_wait)
    return recv_preamble();


  if (!msg) {  // receive into header
    cdebug::instance() << acquire("conn_debug")
                       << "receiving into header"
                       << release();

    assert(hdr_bytes_left > 0 && hdr_bytes_left <= sizeof(hdr));
    int read = recv(sock, (char *)&hdr + sizeof(hdr) - hdr_bytes_left,
                    hdr_bytes_left, 0);

    if (read == -1 || read == 0) {
      cdebug::instance() << acquire("conn_debug") << "receive() ";
      if (read == 0)
        cdebug::instance() << "sock " << sock << "closed." << release();
      else
        cdebug::instance() << "sock " << sock << "couldn't recv, errno "
                           << errno;
      cdebug::instance() << release();
      return kerr_couldnt_receive;
    }
    hdr_bytes_left -= read;

    // do we have the complete header?
    if (hdr_bytes_left == 0) {  // yes, create a new msg
      msg = cmsg_factory::instance().new_msg(hdr);
      if (!msg)
        return kerr_bad_header;

      // set this connection's user
      if (peer_username == "") {
        peer_username = string(hdr.username);
        assert(peer_username != "");
      }
    }
  }

  if (msg) {  // receive into data buffer
    cdebug::instance() << acquire("conn_debug") << "receiving into data"
                       << release();

    uint data_bytes_left;

    uint ret = msg->recv_socket(sock, data_bytes_left);
    if (ret != kerr_success)
      return ret;

    if (data_bytes_left == 0) {  // done reading, process the sucker!
      hdr_bytes_left = sizeof(hdr);

      msg->ntoh();
      bool delmsg = msg->process();

      if (delmsg) {
        delete msg;
      } else
        cdebug::instance() << acquire("conn_debug") 
                           << "did not delete msg after process"
                           << release();
      msg = NULL;
    }
  }
  return kerr_success;
}


void cconn::close()
{
  cdebug::instance() << acquire("conn_debug") << "close(), socket " << sock
                     << release();
  if (status == efailed || status == eclosed)
    return;

  assert(sock >= 0);
#ifdef _WIN32
  closesocket(sock);
#else
  ::close(sock);
#endif

  status = eclosed;
}


/** ctors
 *
 */
cconn::cconn(ulong addr, ushort port, const cmsg &init_msg, 
             const char *username) : peer_username(username)
{
  // debugging
  struct in_addr in;
  in.s_addr = addr;
  char *addr_str = inet_ntoa(in);
  cdebug::instance() << acquire("conn_debug")
                     << "ctor() " << "addr = " << addr_str << " port = "
                     << port
                     << release();

  // initialize members
  assert(peer_username != "");
  status = eopening;
  permanent = false;
  hdr_bytes_left = sizeof(p4_header);

  preamble_bytes_read = 0;
  preamble_buf[0] = '\0';


  // store local copy of message
  msg = cmsg_factory::instance().new_msg(init_msg);
  assert(msg);

  // fill in remote address
  struct sockaddr_in remote;
//  remote.sin_addr.s_addr = htonl(addr);    // why not this?
  remote.sin_addr.s_addr = addr;
  remote.sin_family = AF_INET;
  remote.sin_port = htons(port);


  // create socket and connect
  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    status = efailed;
    return;
  }
  cdebug::instance() << acquire("conn_debug")
                     << "ctor() " << "trying to connect on sock " << sock
                     << release();

#ifndef CONN_MAIN  
  int ret = connect(sock, reinterpret_cast<struct sockaddr *>(&remote),
                    sizeof(remote));
  if (ret == -1) {
    cdebug::instance() << acquire("conn_debug")
                       << "ctor() " << "connect failed, errno " << errno
                       << release();
    status = efailed;
    return;
  }

  cdebug::instance() << acquire("conn_debug")
                     << "ctor() " << "connected...sending preamble"
                     << release();

  // start preamble
  if (send_all(kp4_preamble, sizeof(kp4_preamble)) != kerr_success) {
    cdebug::instance() << acquire("conn_debug")
                       << "ctor() " << "send pre failed, errno " << errno
                       << release();
    status = efailed;
    return;
  }
#endif

  status = epreamble_ack_wait;
  cdebug::instance() << acquire("conn_debug") << "finished ctor" << release();
}


cconn::cconn(int sock, bool failed) : sock(sock), peer_username("")
{
  status = eopening;

  hdr_bytes_left = sizeof(hdr);
  msg = 0;
  permanent = false;

  preamble_bytes_read = 0;
  preamble_buf[0] = '\0';

  if (failed)
    status = efailed;
  else
    status = epreamble_wait;
}


cconn::cconn(const cconn &orig) : peer_username(orig.peer_username),
                                  hdr(orig.hdr)
  
{
  cdebug::instance() << acquire("conn_debug") << "copy ctor()" << release();

  // initialize members
//  assert(peer_username != "");
  sock = orig.sock;
  permanent = orig.permanent;
  hdr_bytes_left = orig.hdr_bytes_left;
  status = orig.status;

  // initialize message stuff
  preamble_bytes_read = orig.preamble_bytes_read;
  string preamble(orig.preamble_buf);
  preamble.copy(preamble_buf, sizeof(preamble_buf));
  preamble_buf[sizeof(preamble_buf) - 1] = '\0';

  if (orig.msg == 0)
    msg = 0;
  else
    msg = cmsg_factory::instance().new_msg(*orig.msg);
}


/*****************************************************************************
 * PRIVATE METHODS
 *****************************************************************************/
uint cconn::send_all(const void *buf, uint data_len) const
{
  const char *cbuf = reinterpret_cast<const char *>(buf);
  uint bytes_left = data_len;

  while (bytes_left > 0) {
    int sent = ::send(sock, cbuf + data_len - bytes_left, bytes_left, 0);
    if (sent == -1 && errno != EINTR) {
      cdebug::instance() << acquire("conn_debug")
                         << "send_all on sock " << sock << " failed.  err "
                         << errno
                         << release();
      return kerr_not_sent;
    }
    bytes_left -= sent;
  }

  verify_socket(sock);
  return kerr_success;
}


uint cconn::recv_preamble()
{
  cdebug::instance() << acquire("conn_debug") << "recv_preamble" << release();

  int read = recv(sock, preamble_buf + preamble_bytes_read,
                  sizeof(preamble_buf) - preamble_bytes_read, 0);
  if (read == -1 || read == 0) {
    cdebug::instance() << acquire("conn_debug") << "recv_preamble() ";
    if (read == 0)
      cdebug::instance() << "sock " << sock << "closed." << release();
    else
      cdebug::instance() << "sock " << sock << "couldn't recv, err " << errno;
    cdebug::instance() << release();
    return kerr_couldnt_receive;
  }
  preamble_bytes_read += read;


  // if we received preamble, send preamble ack
  if (status == epreamble_wait &&
    preamble_bytes_read == sizeof(kp4_preamble)) {

    if (string(preamble_buf) == string(kp4_preamble)) {  // got preamble!
      uint ret = send_all(kp4_preamble_ack, sizeof(kp4_preamble_ack));
      if (ret != kerr_success) {
        status = efailed;
        return ret;
      }

      status = econnected;
      cdebug::instance() << acquire("conn_debug")
                         << "received preamble! sent preamble_ack. connected."
                         << release();

    } else {
      cdebug::instance() << acquire("conn_debug")
                         << "expecting preamble, received " << preamble_buf
                         << release();
      status = efailed;
    }
  }


  // if we received preamble_ack, we're connected!
  if (status == epreamble_ack_wait &&
      preamble_bytes_read == sizeof(kp4_preamble_ack)) {

    if (string(preamble_buf) == string(kp4_preamble_ack)) {
      cdebug::instance() << acquire("conn_debug")
                         << "received preamble_ack! connected."
                         << release();
      status = econnected;

    } else {
      cdebug::instance() << acquire("conn_debug")
                         << "expecting preamble_ack, received " << preamble_buf
                         << release();
      status = efailed;
    }
  }


  // send stored message, if we just connected
  if (status == econnected && msg) {
    uint ret = send(*msg);
    if (ret != kerr_success) {
      status = efailed;
      return ret;
    }
    delete msg;
    msg = 0;
  }

  return kerr_success;
}

#ifdef CONN_MAIN
#include "msg_hello.h"

int main(int argc, char **argv)
{
  int sock;
  ulong addr;
  ushort port;
  cconn conn(addr, port, cmsg_hello(), "raptor_server");

  cconn new_conn(conn);
}

#endif
